;;; -*- syntax: common-lisp; package: clm; base: 10; mode: lisp -*-

(in-package :clm)

;;; fun with special functions
;;;
;;; special functions for use in CLM -- these can be used directly in the run loop (unless [L]=lisp only)
;;; some translated from "Numerical Recipes in Pascal" by Press, Flannery, Teukolsky, Vetterling, 1989.
;;;    Others from "Handbook of Mathematical Functions", Abramowitz and Stegun, 1965.  Still others are
;;;    from "A Numerical Library in C for Scientists and Engineers", Lau, 1995. Useful references
;;;    are Lebedev "Special Functions and Their Applications", Sansone "Orthogonal Functions",
;;;    Watson "A Treatise on the Theory of Bessel Functions", Tolstov "Fourier Series", Walter
;;;    "Wavelets and Other Orthogonal Systems".
;;;    The x-polynomial functions perform a sum of x-functions evaluated at y.
;;;    The x-spectrum functions project a function onto the basis provided by the x-functions (i.e. they
;;;      return a spectrum, the coefficients that when plugged into the x-polynomial routine will recreate the
;;;      original function).
;;;    The x-fm, x-convolve, and x-filter instruments use these to goof around. Actual examples are:
;;;      legendre-convolve, bessel-FM, bessel-convolve, waveshaping
;;;
;;; For sound-related applications of these functions see "Vibration and Sound" Morse.
;;;
;;; Implemented here are:
;;;
;;;     chebyshev-polynomial (sum of a[k]*Tn(x) (or Un(x))), chebyshev-spectrum [L]
;;;     laguerre-polynomial (sum of a[k]*Ln(x)), 
;;;     hermite-polynomial (sum of a[k]*Hn(x)), hermite-spectrum [L]
;;;     legendre-polynomial (sum of a[k]*Pn(x)), legendre-spectrum
;;;
;;; Many of these spectra functions can be speeded up a lot if overlap-add is in use (i.e if the
;;;   x values are the same across calls) -- then the weighted polynomials can be precalculated and only
;;;   a dot product+scaling is needed to get each successive spectrum.
;;;
;;; this sequence mimics sin/cos | fft | convolve | fft-filter using various orthogonal polynomials
;;;
;;;     bes-j0 (x)      J0(x) for any real x
;;;     bes-j1 (x)	J1(x) for any real x
;;;     bes-jn (nn x)	Jn(x) for any integer n, real x
;;;     bes-y0 (x)	Y0(x)
;;;     bes-y1 (x)	Y1(x)
;;;     bes-yn (n x)	Yn(x) for any integer n, real x
;;;     bes-i0 (x)	I0(x)
;;;     bes-i1 (x)	I1(x)
;;;     bes-in (n x)	In(x) for any integer n, real x
;;;     bes-k0 (x)	K0(x)
;;;     bes-k1 (x)	K1(x)
;;;     bes-kn (n x)	Kn(x) for any integer n, real x
;;;     bes-j (et al - any real order, any x)
;;;     bessel-spectrum [L] bessel-polynomial [L]
;;;     nes-zero
;;;
;;;     Ai and Bi       Airy functions (and some related fractional-order Bessel functions: Jp3 Jm3 Ip3 Im3)
;;;     Si and Ci       sine and cosine integrals
;;;
;;;     various gamma function related stuff (just as macros)
;;;     spherical bessel functions [L]
;;;
;;; the term "convolve" is used loosely here to refer to an analog to convolution as per fft --
;;; i.e. take spectrum, multiply, invert.  "real" convolution is defined differently sometimes.
;;; See Zayed "Handbook of Function Transformation" -- the real point here was to experiment
;;; with other kinds of spectra.

(export '(gammln gamma
	  chebyshev-polynomial laguerre-polynomial hermite-polynomial
	  legendre-polynomial bessel-polynomial
	  legendre-spectrum bessel-spectrum
	  chebyshev-spectrum hermite-spectrum
	  bes-j0 bes-j1 bes-jn
	  bes-y0 bes-y1 bes-yn
	  bes-i0 bes-i1 bes-in
	  bes-k0 bes-k1 bes-kn bes-zero
	  Ai Bi Sint Cint Jp3 Jm3 Ip3 Im3
	  beta gser gcf erfc betai betacf erf gammq gammp binomial
	  spher-bes-jn spher-bes-yn spher-bes-in spher-bes-kn))


;;; -------- GAMMA, GAMMLN --------
;;;
;;; normally we replace calls on gamma (i.e. factorial essentially) with gammln -- since the
;;;   formulas usually divide one gamma by another (which can mean enormous numbers in the
;;;   division), we try to use (exp (- (gammln num) (gammln den))) instead to avoid overflows.
;;; gammln from "Numerical Recipes" p176 (p214 new ed).

(defun gammln (xx)			;Ln(gamma(xx)), xx>0 
  (let* ((stp 2.5066282746310005d0)
	 (x xx)
	 (tmp (+ x 5.5))
	 (tmp1 (- tmp (* (+ x 0.5) (log tmp))))
	 (ser (+ 1.000000000190015
		 (/ 76.18009172947146 (+ x 1.0))
		 (/ -86.50532032941677 (+ x 2.0))
		 (/ 24.01409824083091 (+ x 3.0))
		 (/ -1.231739572450155 (+ x 4))
		 (/ 0.1208650973866179e-2 (+ x 5.0))
		 (/ -0.5395239384953e-5 (+ x 6.0)))))
    (- (log (/ (* stp ser) x)) tmp1)))

(def-clm-fun 'gammln #'(lambda (var x) (package-op '<gammln> var x :clm-real)))

(defmacro <gammln> (result xx)
  (let ((lcr (lc (second result)))
	(lcxx (lc-num-ref xx)))
    (format *c-file* "  {double stp,x,tmp,tmp1,ser; x=~A-1.0;~80,1T/* (gammln ~A) */~%   ~
                         stp=2.50662827465; tmp=x+5.5; tmp1=((x+0.5)*log(tmp))-tmp;~%   ~
                         ser=1.0+(76.18009173/(x+1.0))-(86.50532033/(x+2.0))+(24.01409822/(x+3.0))-~%      ~
                             (1.231739516/(x+4.0))+(0.120858003e-2/(x+5.0))-(0.536382e-5/(x+6.0));~%   ~
                         ~A=tmp1+log(stp*ser);}~%"
	    lcxx (clean-arg xx) lcr)
    nil))

(defmacro gamma (ux)			;gamma(x)
  `(let ((x ,ux))
     (if (plusp x)
	 (if (>= x 1.0)
	     (exp (gammln x))
	   (/ (exp (gammln (+ x 1.0))) x))
       (/ (* pi (exp (- (gammln (- 1.0 x))))) (sin (* pi (- 1.0 x)))))))




;;; -------- CHEBYSHEV POLYNOMIAL --------
;;;
;;; A&S p782 22.7.4 and 22.7.5 Tn and Un
;;;
;;; orthogonal with weight (1-x^2)^-.5 on the interval -1..1 (A&S 22.2.4)
;;; |Tn(x)|<=1 when -1<=x<-1, |Un(x)|<=n+1 on same interval A&S 22.14.4 22.14.6

(defun chebyshev-polynomial (a x &optional kind lim)	
  ;; evaluate the sum of the Chebyshev polynomials (coeffs in a) at x
  ;;  similar to make-waveshape-table which runs through -1<=x<=1 internally,
  ;;  but intended to be parallel to the polynomial unit generator --
  ;; (polynomial (get-chebyshev-coefficients harms-and-amps) x) is equivalent to 
  ;; (chebyshev-polynomial harm-amps x)
  (let ((n (or lim (1- (length a))))
	(knd (or kind 1)))
    (if (= n 0) (aref a 0)
      (let* ((r (* knd x))
	     (s 1.0)
	     (h 0.0)
	     (sum (aref a 0)))
	(loop for k from 1 below n do
	  (setf h r)
	  (incf sum (* r (aref a k)))
	  (setf r (- (* 2 r x) s))
	  (setf s h))
	(+ sum (* r (aref a n)))))))

(def-clm-fun 'chebyshev-polynomial #'(lambda (var x) (package-op '<chebyshev-polynomial> var x :clm-real)))

(defmacro <chebyshev-polynomial> (result a ux &optional kind lim)
  (let ((lcr (lc (second result)))
	(lcx (lc-num-ref ux)))
    (format *c-file* "  {int k,n; double h,r,s,x,sum; double *arr;~80,1T/* (chebyshev-polynomial ~A ~A ~A ~A) */~%  ~
                         arr=(double *)(clm_double+~A); n=~A; x=~A;~%   ~
	                 if (n == 0) ~A=arr[0];~%   ~
                         else {r=x~A; s=1.0; sum=arr[0];~%         ~
                               for (k=1;k<n;k++) {h=r; sum+=r*arr[k]; r=(2*r*x)-s; s=h;}~%        ~
                               ~A=sum+(r*arr[n]);}}~%" 
	    a ux lim kind
	    (array-rblock a) (if lim (lc-num-ref lim) (array-size a)) lcx
	    lcr 
	    (if (and kind (or (not (numberp kind)) (= kind 2))) (format nil "*~A" (lc-num-ref kind)) "")
	    lcr))
    nil)

#|
;;; standard waveshaping:
(definstrument waveshaping (beg dur amp freq a)
  (let* ((st (floor (* *srate* beg)))
	 (nd (+ st (floor (* *srate* dur))))
	 (rate (/ (* 2 pi freq) *srate*))
	 (ph 0.0))
    (run
     (loop for i from st to nd do
       (outa i (* amp (chebyshev-polynomial a (cos ph))))
       (incf ph rate)))))

;;; (with-sound () (waveshaping 0 1 .1 440.0 (make-array 5 :initial-contents '(0.0 1.0 1.0 .5 .25))))
|#

(defun chebyshev-spectrum (lim Fn samps An Pn kind)
  (flet ((chebyshev-polynomials (x f knd &optional lim)
	   (let ((n (or lim (1- (length f)))))
	     (let* ((r (* knd x))
		    (s 1.0)
		    (h 0.0))
	       (setf (aref f 0) 1.0)
	       (setf (aref f 1) r)
	       (loop for k from 1 below n do
		 (setf h r)
		 (setf r (- (* 2 r x) s))
		 (setf (aref f (+ k 1)) r)
		 (setf s h))))))
    (let* ((x -1.0d0)
	   (rate (/ 2.0 samps)))
      (loop for i from 0 to lim do (setf (aref An i) 0.0))
      (loop for k from 0 below samps do
	(let* ((fx (aref Fn k))
	       (wx (if (= kind 1)
		       (if (= (abs x) 1.0)
			   (log samps 2)         ;just a guess
			 (/ 1.0 (sqrt (- 1.0 (* x x)))))
		     (sqrt (- 1.0 (* x x)))))
	       (wfx (* wx fx)))
	  (chebyshev-polynomials x Pn kind lim)
	  (loop for i from 0 to lim do 
	    (incf (aref An i) (* wfx (aref Pn i)))))
	(incf x rate))	
      (let ((wc (/ 2.0 (* pi samps))))
	(setf (aref An 0) (* (aref An 0) wc (if (= kind 1) 2.0 1.0)))
	(loop for i from 1 to lim do	           
	  (setf (aref An i) (* (aref An i) wc))))
      An)))

#|
;;; test function
(defun cs (&optional (n 256) (lim 200))
  (setf Fn (make-array n :initial-element 0.0))
  (setf An (make-array n :initial-element 0.0))
  (setf Pn (make-array n :initial-element 0.0))
  (setf Xn (make-array n :initial-element 0.0))
  (loop for i from 0 below n do (setf (aref Fn i) (* i (/ 1.0 n))))
  (chebyshev-spectrum lim Fn (1- n) An Pn 1)
  (loop for i from 0 below n do (setf (aref Xn i) (chebyshev-polynomial An (- (* i (/ 2.0 n)) 1.0) 1 lim)))
  Xn)
|#


;;; -------- HERMITE POLYNOMIAL --------
;;;
;;; A&S p782 22.7.13 Hn
;;;
;;; the functions (2^n n! pi^.5)^.5  e^(-x^2) Hn(x) form an orthonormal basis on the interval -inf..inf (A&S 22.2.14)
;;;   (Lebedev "Special Functions" p 66 -- see also sec 4.15 p68 and Sansone p308 and p368)
;;;   they are eigenfunctions of the Fourier transform (c.f. Walter "Wavelets and Other Orthogonal Systems" p22)
;;; |Hn(x)|<e^((x^2)/2) * 1.086435 * 2^(n/2) * sqrt(n!) A&S 22.14.17
;;; (so coeff weight is 1/(2^n n! (sqrt pi)), function weight is e^(-x^2))

(defun hermite-polynomial (a x &optional lim)	
  (let ((n (or lim (1- (length a)))))
    (if (= n 0) (aref a 0)
      (let* ((r (* 2 x))
	     (s 1.0)
	     (h 0.0)
	     (sum (aref a 0)))
	(loop for k from 1 below n and k2 from 2 by 2 do
	  (setf h r)
	  (incf sum (* r (aref a k)))
	  (setf r (- (* 2 r x) (* k2 s)))
	  (setf s h))
	(+ sum (* r (aref a n)))))))

(defun hermite-spectrum (lim Fn samps An Pn)
  (flet ((hermite-polynomials (x f &optional lim)
	   (let* ((n (or lim (1- (length f))))
		  (r (* 2 x))
		  (s 1.0)
		  (h 0.0))
	     (setf (aref f 0) 1.0)
	     (setf (aref f 1) r)
	     (loop for k from 1 below n and k2 from 2 by 2 do
	       (setf h r)
	       (setf r (- (* 2 r x) (* k2 s)))
	       (setf (aref f (+ k 1)) r)
	       (setf s h)))))
    (let* ((x -10.0d0)			;assume e^(- x^2) will dominate as weight function
	   (rate (/ 20.0 samps)))
      (loop for i from 0 to lim do (setf (aref An i) 0.0))
      (loop for k from 0 below samps do
	(let* ((fx (aref Fn k))
	       (wx (exp (- (* x x))))
	       (xfx (* wx fx)))
	  (hermite-polynomials x Pn lim)
	  (loop for i from 0 to lim do 
	    (incf (aref An i) (* xfx (aref Pn i)))))
	(incf x rate))			
      (let ((scaler (sqrt pi)))
	(setf (aref An 0) (/ (aref An 0) scaler))
	(loop for i from 1 to lim do	             
	  (setf (aref An i) (/ (aref An i) scaler))
	  (setf scaler (* scaler (* 2 i)))))
      An)))

(def-clm-fun 'hermite-polynomial #'(lambda (var x) (package-op '<hermite-polynomial> var x :clm-real)))

(defmacro <hermite-polynomial> (result a ux &optional lim)
  (let ((lcr (lc (second result)))
	(lcx (lc-num-ref ux)))
    (format *c-file* "  {int k,n,k2; double h,r,s,x,sum; double *arr;~80,1T/* (hermite-polynomial ~A ~A ~A) */~%  ~
                         arr=(double *)(clm_double + ~A); n=~A; x=~A;~%   ~
	                 if (n == 0) ~A=arr[0];~%   ~
                         else {r=2*x; s=1.0; sum=arr[0];~%         ~
                               for (k=1,k2=2;k<n;k++,k2+=2) {h=r; sum+=r*arr[k]; r=(2*r*x)-(k2*s); s=h;}~%        ~
                               ~A=sum+(r*arr[n]);}}~%" 
	    a ux lim
	    (array-rblock a) (if lim (lc-num-ref lim) (array-size a)) lcx
	    lcr 
	    lcr))
    nil)



;;; -------- LAGUERRE POLYNOMIAL --------
;;;
;;; A&S p782 22.7.12 Ln
;;; |Ln(x)|<=e^(x/2) A&S 22.14.12 and the weight is dominated by e^(-x) (A&S 22.2.12), so I think
;;;   it is ok to map 0..inf as 0..100 (0..10 is probably even safe, despite the x^a term)

(defun laguerre-polynomial (a x &optional (alpha 0.0) lim);odd that "Numerical Library" has the wrong recursion here (p615)
  (let ((n (or lim (1- (length a)))))
    (if (= n 0) (aref a 0)
      (let* ((r (- (+ alpha 1.0) x))
	     (s 1.0)
	     (h 0.0)
	     (sum (aref a 0)))
	(loop for k from 1 below n do
	  (setf h r)
	  (incf sum (* r (aref a k)))
	  (setf r (/ (- (* r (- (+ (* 2 k) 1 alpha) x)) (* s (+ k alpha))) (+ k 1)))
	  (setf s h))
	(+ sum (* r (aref a n)))))))

(def-clm-fun 'laguerre-polynomial #'(lambda (var x) (package-op '<laguerre-polynomial> var x :clm-real)))

(defmacro <laguerre-polynomial> (result a ux &optional alpha lim)
  (let ((lcr (lc (second result)))
	(lcx (lc-num-ref ux))
	(lcalpha (if alpha (lc-num-ref alpha) "")))
    (format *c-file* "  {int k,n,k2; double h,r,s,x,sum; double *arr;~80,1T/* (laguerre-polynomial ~A ~A ~A ~A) */~%  ~
                         arr=(double *)(clm_double+~A); n=~A; x=~A;~%   ~
	                 if (n == 0) ~A=arr[0];~%   ~
                         else {r=1.0~A-x; s=1.0; sum=arr[0];~%         ~
                               for (k=1,k2=1;k<n;k++,k2+=2)~%           ~
                                  {h=r; sum+=r*arr[k]; r=(-((r*(x-(k2~A))+(s*(k~A)))/(k+1))); s=h;}~%        ~
                               ~A=sum+(r*arr[n]);}}~%" 
	    a ux lim alpha
	    (array-rblock a) (if lim (lc-num-ref lim) (array-size a)) lcx
	    lcr lcalpha lcalpha lcalpha lcr))
  nil)


;;; -------- LEGENDRE POLYNOMIAL --------
;;;
;;; also known as spherical polynomials, A&S p782 22.7.10 Pn, A&S 22.2.10
;;;
;;; the functions (n+.5)^.5 Pn(x) form an orthonormal basis on the interval -1..1 (Lebedev p 51) (see also sec 4.7 p53)
;;;   for a function on [a..b] mapped into [-1..1], the weight becomes (2n+1)/(b-a) (Sansone "Orthogonal Functions" p245)
;;; |Ln(x)|<=1 A&S 22.14.7
;;; for waves on the surface of a sphere (Butkov 9.8) and a whirling string (Morse p118)

(defun legendre-polynomial (a x &optional lim)
  (let ((n (or lim (1- (length a)))))
    (if (= n 0) (aref a 0)
      (let* ((r x)
	     (s 1.0)
	     (h 0.0)
	     (sum (aref a 0)))
	(loop for k from 1 below n do
	  (setf h r)
	  (incf sum (* r (aref a k)))
	  (setf r (/ (- (* r x (+ (* 2 k) 1))  (* s k)) (+ k 1)))
	  (setf s h))
	(+ sum (* r (aref a n)))))))

(def-clm-fun 'legendre-polynomial #'(lambda (var x) (package-op '<legendre-polynomial> var x :clm-real)))

(defmacro <legendre-polynomial> (result a ux &optional lim)
  (let ((lcr (lc (second result)))
	(lcx (lc-num-ref ux)))
    (format *c-file* "  {int k,n,k2; double h,r,s,x,sum; double *arr;~80,1T/* (legendre-polynomial ~A ~A ~A) */~%  ~
                         arr=(double *)(clm_double+~A); n=~A; x=~A;~%   ~
	                 if (n == 0) ~A=arr[0];~%   ~
                         else {r=x; s=1.0; sum=arr[0];~%         ~
                               for (k=1,k2=3;k<n;k++,k2+=2) {h=r; sum+=r*arr[k]; r=(((r*x*k2)-(s*k))/(k+1)); s=h;}~%        ~
                               ~A=sum+(r*arr[n]);}}~%" 
	    a ux lim
	    (array-rblock a) 
	    (if lim (lc-num-ref lim) (array-size a))
	    lcx
	    lcr 
	    lcr))
  nil)

(defun legendre-spectrum (lim Fn samps An Pn)
  ;; expand function in Fn as a sum of Legendre Functions.
  ;; coefficients (spectrum) returned in An, Pn is temporary working space -- both are arrays (1+ lim) in size
  ;; lim is how many Legendre functions to use in the expansion -- should be a lot!
  ;; samps is how many samples to read
  ;; The run translation to C constrains our choice of arguments here (no internally allocated arrays, etc).
  (flet ((legendre-polynomials (x f &optional lim)
	   (let ((n (or lim (1- (length f)))))
	     (let* ((r x)
		    (s 1.0d0)
		    (h 0.0))
	       (setf (aref f 0) 1.0)
	       (setf (aref f 1) x)
	       (loop for k from 1 below n do
		 (setf h r)
		 (setf r (/ (- (* r x (+ (* 2 k) 1))  (* s k)) (+ k 1)))
		 (setf (aref f (+ k 1)) r)
		 (setf s h))))))
     (let* ((x -1.0d0)
	    (rate (/ 2.0 samps)))
       (loop for i from 0 to lim do (setf (aref An i) 0.0))
       (loop for k from 0 below samps do
	 (let ((fx (aref Fn k)))
	   (legendre-polynomials x Pn lim)           ;Pn(x) for each n
	   (loop for i from 0 to lim do 
	     (incf (aref An i) (* fx (aref Pn i))))) ;f(x)*Pn(x) summed as we move through the sound
	 (incf x rate))			             ;-1<=x<=1 (interval of orthonormality for Pn)
       An)))


(defun continued-fraction (lim a b)	;lim=top index, a=numerators, b=denominators in expression of continued fraction
  (let ((d 0.0))
    (loop for i from lim downto 1 do (setf d (/ (aref a i) (+ (aref b i) d))))
    (+ (aref b 0) d)))



;;; -------- BESSEL FUNCTIONS --------
;;;
;;; The basic functions here are translated from "Numerical Recipes" which appear to me
;;;   to be better than those of "Numerical Library" -- the latter even has a bug in bessj1!

(defun bes-j0 (x)				;returns J0(x) for any real x
  (if (< (abs x) 8.0)			;direct rational function fit
      (let* ((y (* x x))
	     (ans1 (+ 57568490574.0
		      (* y (+ -13362590354.0 
			      (* y  (+ 651619640.7
				       (* y (+ -11214424.18 
					       (* y (+ 77392.33017
						       (* y -184.9052456)))))))))))
	     (ans2 (+ 57568490411.0 
		      (* y (+ 1029532985.0 
			      (* y (+ 9494680.718
				      (* y (+ 59272.64853
					      (* y (+ 267.8532712 y)))))))))))
	(/ ans1 ans2))
    (let* ((ax (abs x))
	   (z (/ 8.0 ax))
	   (y (* z z))
	   (xx (- ax 0.785398164))
	   (ans1 (+ 1.0 
		    (* y (+ -0.1098628627e-2 
			    (* y (+ 0.2734510407e-4
				    (* y (+ -0.2073370639e-5
					    (* y 0.2093887211e-6)))))))))
	   (ans2 (+ -0.1562499995e-1
		    (* y (+ 0.1430488765e-3
			    (* y (+ -0.6911147651e-5
				    (* y (+ 0.7621095161e-6
					    (* y -0.934945152e-7))))))))))
      (* (sqrt (/ 0.636619772 ax))
	 (- (* (cos xx) ans1)
	    (* z (sin xx) ans2))))))

(def-clm-fun 'bes-j0 #'(lambda (var x) (package-op '<bes-j0> var x :clm-real)))

(defmacro <bes-j0> (result ux)
  (let ((lcx (lc-num-ref ux))
	(lcr (lc (second result))))
    (format *c-file* 
            "  {double x,y,ans1,ans2,z,xx,ax; x=~A;~80,1T/* (bes-j0 ~A) */~%   ~
                if ((x < 8.0) && (x > -8.0))~%     ~
                  {y=x*x;~%      ~
                   ans1=57568490574.0+(y*(-13362590354.0+(y*(651619640.7+~%      ~
                                        (y*(-11214424.18+(y*(77392.33017+(y* -184.9052456)))))))));~%      ~
                   ans2=57568490411.0+(y*(1029532985.0+(y*(9494680.718+(y*(59272.64853+(y*(267.8532712+y))))))));~%      ~
                   ~A = (ans1/ans2);}~%   ~
                else~%     ~
                  {ax=x; if (ax<0.0) ax= -x; z=8.0/ax; y=z*z; xx=ax-0.785398164;~%      ~
	           ans1=1.0+(y*(-0.1098628627e-2+(y*(0.2734510407e-4+(y*(-0.2073370639e-5+(y*0.2093887211e-6)))))));~%      ~
	           ans2=-0.1562499995e-1+(y*(0.1430488765e-3+~%      ~
                                            (y*(-0.6911147651e-5+(y*(0.7621095161e-6+(y*-0.934945152e-7)))))));~%      ~
                   ~A = (sqrt(0.636619772/ax))*((cos(xx)*ans1)-(z*sin(xx)*ans2));}}~%"
	    lcx ux lcr lcr)
    nil))

(defun bes-j1 (x)				;returns J1(x) for any real x
  (if (< (abs x) 8.0)
      (let* ((y (* x x))
	     (ans1 (* x 
		      (+ 72362614232.0
			 (* y (+ -7895059235.0
				 (* y (+ 242396853.1
					 (* y (+ -2972611.439
						 (* y (+ 15704.48260
							 (* y -30.16036606))))))))))))
	     (ans2 (+ 144725228442.0 
		      (* y (+ 2300535178.0 
			      (* y (+ 18583304.74
				      (* y (+ 99447.43394
					      (* y (+ 376.9991397 y)))))))))))
	(/ ans1 ans2))
    (let* ((ax (abs x))
	   (z (/ 8.0 ax))
	   (y (* z z))
	   (xx (- ax 2.356194491))
	   (ans1 (+ 1.0
		    (* y (+ 0.183105e-2
			    (* y (+ -0.3516396496e-4
				    (* y (+ 0.2457520174e-5
					    (* y -0.240337019e-6)))))))))
	   (ans2 (+ 0.04687499995
		    (* y (+ -0.2002690873e-3
			    (* y (+ 0.8449199096e-5
				    (* y (+ -0.88228987e-6
					    (* y 0.105787412e-6))))))))))
      (* (signum x)
	 (sqrt (/ 0.636619772 ax))
	 (- (* (cos xx) ans1)
	    (* z (sin xx) ans2))))))

(def-clm-fun 'bes-j1 #'(lambda (var x) (package-op '<bes-j1> var x :clm-real)))

(defmacro <bes-j1> (result ux)
  (let ((lcx (lc-num-ref ux))
	(lcr (lc (second result))))
    (format *c-file* 
            "  {double x,y,ans1,ans2,z,xx,ax; x=~A;~80,1T/* (bes-j1 ~A) */~%   ~
                if ((x < 8.0) && (x > -8.0))~%     ~
                  {y=x*x;~%      ~
                   ans1=x*(72362614232.0+(y*(-7895059235.0+(y*(242396853.1+~%      ~
                                           (y*(-2972611.439+(y*(15704.48260+(y* -30.16036606))))))))));~%     ~
                   ans2=144725228442.0+(y*(2300535178.0+(y*(18583304.74+(y*(99447.43394+(y*(376.9991397+y))))))));~%     ~
                   ~A = (ans1/ans2);}~%   ~
                else~%     ~
                  {ax=x; if (ax<0.0) ax= -x; z=8.0/ax; y=z*z; xx=ax-2.356194491;~%     ~
                   ans1=1.0+(y*(0.183105e-2+(y*(-0.3516396496e-4+(y*(0.2457520174e-5+(y* -0.240337019e-6)))))));~%     ~
                   ans2=0.04687499995+(y*(-0.2002690873e-3+(y*(0.8449199096e-5+(y*(-0.88228987e-6+(y*0.105787412e-6)))))));~%     ~
                   ~A = (sqrt(0.636619772/ax))*((cos(xx)*ans1)-(z*sin(xx)*ans2));~%     ~
                   if (x < 0.0) ~A = -~A;}}~%"
	    lcx ux lcr lcr lcr lcr)
    nil))


(defmacro bes-jn (unn ux)				;return Jn(x) for any integer n, real x
  `(let ((nn ,unn) (x ,ux))
     (let* ((n (floor (abs nn)))
	    (besn (if (= n 0) (bes-j0 x)
		    (if (= n 1) (bes-j1 x)
		      (if (zerop x) 0.0
			(let ((iacc 40)	;make iacc larger to increase accuracy
			      (ans 0.0)
			      (bigno 1.0e10)
			      (bigni 1.0e-10))
			  (if (> (abs x) n) ;can use upward recurrence from J0 and J1
			      (do ((tox (/ 2.0 (abs x)))
				   (bjm (bes-j0 (abs x)))
				   (bj (bes-j1 (abs x)))
				   (j 1 (+ j 1))
				   (bjp 0.0))
				  ((= j n) (setf ans bj))
				(setf bjp (- (* j tox bj) bjm))
				(setf bjm bj)
				(setf bj bjp))
			    (let ((tox (/ 2.0 (abs x))) ;else use downward recurrence from even value (m)
				  (m (* 2 (floor ;DIV in Pascal, we hope...
					   (+ n (sqrt (* iacc n))) 2)))
				  (jsum 0.0) ;alternate 0 and 1 -- when 1, accumulate even terms in sum
				  (bjm 0.0)
				  (sum 0.0)
				  (bjp 0.0)
				  (bj 1.0))
			      (do ((j m (- j 1))) ;the downward recurrence
				  ((= j 0))
				(setf bjm (- (* j tox bj) bjp))
				(setf bjp bj)
				(setf bj bjm)
				(when (> (abs bj) bigno) ;renormalize (may not be necessary in common lisp)
				  (setf bj (* bj bigni))
				  (setf bjp (* bjp bigni))
				  (setf ans (* ans bigni))
				  (setf sum (* sum bigni)))
				(if (/= 0 jsum) (incf sum bj))
				(setf jsum (- 1 jsum))
				(if (= j n) (setf ans bjp)))
			      (setf sum (- (* 2.0 sum) bj))
			      (setf ans (/ ans sum))))
			  (if (and (minusp x) (oddp n)) (- ans) ans)))))))
       (if (and (minusp nn)
		(oddp nn))
	   (- besn)
	 besn))))


(defmacro bes-y0 (ux)				;Bessel function Y0(x)
  `(let ((x ,ux))
     (if (< x 8.0)
	 (let* ((y (* x x))
		(ans1 (+ -2957821389.0 
			 (* y (+ 7062834065.0
				 (* y (+ -512359803.6
					 (* y (+ 10879881.29
						 (* y (+ -86327.92757
							 (* y 228.4622733)))))))))))
		(ans2 (+ 40076544269.0
			 (* y (+ 745249964.8
				 (* y (+ 7189466.438
					 (* y (+ 47447.26470
						 (* y (+ 226.1030244 y)))))))))))
	   (+ (/ ans1 ans2) (* 0.636619772 (bes-j0 x) (log x))))
       (let* ((z (/ 8.0 x))
	      (y (* z z))
	      (xx (- x 0.785398164))
	      (ans1 (+ 1.0
		       (* y (+ -0.1098628627e-2
			       (* y (+ 0.2734510407e-4
				       (* y (+ -0.2073370639e-5
					       (* y 0.2093887211e-6)))))))))
	      (ans2 (+ -0.1562499995e-1
		       (* y (+ 0.1430488765e-3
			       (* y (+ -0.6911147651e-5
				       (* y (+ 0.7621095161e-6
					       (* y -0.934945152e-7)))))))))
	      (ans (+ (* (sin xx) ans1) (* z (cos xx) ans2))))
	 (* (sqrt (/ 0.636619772 x)) ans)))))
	     

(defmacro bes-y1 (ux)				;Bessel function Y1(x)
  `(let ((x ,ux))
     (if (< x 8.0)
	 (let* ((y (* x x))
		(ans1 (* x (+ -0.4900604943e13
			      (* y (+ 0.1275274390e13
				      (* y (+ -0.5153438139e11
					      (* y (+ 0.7349264551e9
						      (* y (+ -0.4237922726e7
							      (* y 0.8511937935e4))))))))))))
		(ans2 (+ 0.2499580570e14
			 (* y (+ 0.4244419664e12
				 (* y (+ 0.3733650367e10
					 (* y (+ 0.2245904002e8
						 (* y (+ 0.1020426050e6
							 (* y (+ 0.3549632885e3 y)))))))))))))
	   (+ (/ ans1 ans2) (* 0.636619772 (- (* (bes-j1 x) (log x)) (/ 1.0 x)))))
       (let* ((z (/ 8.0 x))
	      (y (* z z))
	      (xx (- x 2.356194491))
	      (ans1 (+ 1.0 
		       (* y (+ 0.183105e-2
			       (* y (+ -0.3516396496e-4
				       (* y (+ 0.2457520174e-5
					       (* y -0.240337019e-6)))))))))
	      (ans2 (+ 0.04687499995
		       (* y (+ -0.200269087e-3
			       (* y (+ 0.8449199096e-5
				       (* y (+ -0.88228987e-6
					       (* y 0.105787412e-6))))))))))
	 (* (sqrt (/ 0.636619772 x)) (+ (* (sin xx) ans1) (* z (cos xx) ans2)))))))
	     

(defmacro bes-yn (un ux)				;return Yn(x) for any integer n, real x
  `(let ((n ,un) (x ,ux))
     (if (= n 0) (bes-y0 x)
       (if (= n 1) (bes-y1 x)
	 (do* ((tox (/ 2.0 x))
	       (byp 0.0)
	       (by (bes-y1 x))
	       (bym (bes-y0 x))
	       (j 1 (+ j 1)))
	     ((= j n) by)
	   (setf byp (- (* j tox by) bym))
	   (setf bym by)
	   (setf by byp))))))

#|
;;; moved to mus.lisp 1-Feb-96
(defun bes-i0 (x)			;I0(x)
  (if (< (abs x) 3.75)
      (let* ((y (expt (/ x 3.75) 2)))
	(+ 1.0
	   (* y (+ 3.5156229
		   (* y (+ 3.0899424
			   (* y (+ 1.2067492
				   (* y (+ 0.2659732
					   (* y (+ 0.360768e-1
						   (* y 0.45813e-2)))))))))))))
    (let* ((ax (abs x))
	   (y (/ 3.75 ax)))
      (* (/ (exp ax) (sqrt ax)) 
	 (+ 0.39894228
	    (* y (+ 0.1328592e-1
		    (* y (+ 0.225319e-2
			    (* y (+ -0.157565e-2
				    (* y (+ 0.916281e-2
					    (* y (+ -0.2057706e-1
						    (* y (+ 0.2635537e-1
							    (* y (+ -0.1647633e-1
								    (* y 0.392377e-2))))))))))))))))))))
|#

(def-clm-fun 'bes-i0 #'(lambda (var x) (package-op '<bes-i0> var x :clm-real)))

(defmacro <bes-i0> (result ux)
  (let ((lcx (lc-num-ref ux))
	(lcr (lc (second result))))
    (format *c-file* 
            "  {double x,y,ax; x=~A;~80,1T/* (bes-i0 ~A) */~%   ~
                if ((x<3.75) && (x>-3.75))~%     ~
                  {y=(x/3.75)*(x/3.75);~%      ~
                   ~A=1.0+(y*(3.5156229+(y*(3.0899424+(y*(1.2067492+(y*(0.2659732+(y*(0.360768e-1+(y*0.45813e-2)))))))))));}~%   ~
                else {ax=x; if (ax<0.0) ax= -x; y=(ax/3.75);~%      ~
                      ~A=(exp(ax)/sqrt(ax))*(0.39894228+(y*(0.1328592e-1+(y*(0.225319e-2+(y*(-0.157565e-2+~%         ~
                           (y*(0.916281e-2+(y*(-0.2057706e-1+(y*(0.2635537e-1+(y*(-0.1647633e-1+(y*0.392377e-2))))))))))))))));}}~%"
	    lcx ux lcr lcr)
    nil))

(defun bes-i1 (x)				;I1(x)
  (if (< (abs x) 3.75)
      (let* ((y (expt (/ x 3.75) 2)))
	(* x (+ 0.5
		(* y (+ 0.87890594
			(* y (+ 0.51498869
				(* y (+ 0.15084934
					(* y (+ 0.2658733e-1
						(* y (+ 0.301532e-2
							(* y 0.32411e-3))))))))))))))
    (let* ((ax (abs x))
	   (y (/ 3.75 ax))
	   (ans1 (+ 0.2282967e-1
		    (* y (+ -0.2895312e-1
			    (* y (+ 0.1787654e-1 
				    (* y -0.420059e-2)))))))
	   (ans2 (+ 0.39894228
		    (* y (+ -0.3988024e-1
			    (* y (+ -0.362018e-2
				    (* y (+ 0.163801e-2
					    (* y (+ -0.1031555e-1 (* y ans1)))))))))))
	   (sign (if (minusp x) -1.0 1.0)))
      (* (/ (exp ax) (sqrt ax)) ans2 sign))))

(def-clm-fun 'bes-i1 #'(lambda (var x) (package-op '<bes-i1> var x :clm-real)))

(defmacro <bes-i1> (result ux)
  (let ((lcx (lc-num-ref ux))
	(lcr (lc (second result))))
    (format *c-file* 
            "  {double x,y,ax,ans1,ans2,sign; x=~A;~80,1T/* (bes-i1 ~A) */~%   ~
                if ((x<3.75) && (x>-3.75))~%     ~
                  {y=(x/3.75)*(x/3.75);~%      ~
                   ~A=x*(0.5+(y*(0.87890594+(y*(0.51498869+(y*(0.15084934+~%       ~
                                (y*(0.2658733e-1+(y*(0.301532e-2+(y*0.32411e-3))))))))))));}~%   ~
                else {ax=x; if (ax<0.0) ax= -x; y=(ax/3.75);~%      ~
                      ans1=0.2282967e-1+(y*(-0.2895312e-1+(y*(0.1787654e-1+(y*-0.420059e-2)))));~%      ~
                      ans2=0.39894228+(y*(-0.3988024e-1+(y*(-0.362018e-2+(y*(0.163801e-2+(y*(-0.1031555e-1+(y*ans1)))))))));~%      ~
                      if (x<0.0) sign= -1.0; else sign=1.0;~%      ~
                      ~A = (exp(ax)/sqrt(ax))*ans2*sign;}}~%"
	    lcx ux lcr lcr)
    nil))


(defmacro bes-in (un ux)			;return In(x) for any integer n, real x
  `(let ((n ,un) (x ,ux))
     (if (= n 0) (bes-i0 x)
       (if (= n 1) (bes-i1 x)
	 (if (zerop x) 0.0
	   (let* ((iacc 40)
		  (bigno 1.0e10)
		  (bigni 1.0e-10)
		  (ans 0.0)
		  (tox (/ 2.0 (abs x)))
		  (bip 0.0)
		  (bi 1.0)
		  (m (* 2 (+ n (truncate (sqrt (* iacc n))))))
		  (bim 0.0))
	     (loop for j from m downto 1 do
	       (setf bim (+ bip (* j tox bi)))
	       (setf bip bi)
	       (setf bi bim)
	       (when (> (abs bi) bigno)
		 (setf ans (* ans bigni))
		 (setf bi (* bi bigni))
		 (setf bip (* bip bigni)))
	       (if (= j n) (setf ans bip)))
	     (if (and (minusp x) (oddp n)) (setf ans (- ans)))
	     (* ans (/ (bes-i0 x) bi))))))))

		

(defmacro bes-k0 (ux)				;K0(x)
  `(let ((x ,ux))
     (if (<= x 2.0)
	 (let* ((y (* x (/ x 4.0))))
	   (+ (* (- (log (/ x 2.0))) (bes-i0 x)) -0.57721566
	      (* y (+ 0.42278420
		      (* y (+ 0.23069756
			      (* y (+ 0.3488590e-1
				      (* y (+ 0.262698e-2
					      (* y (+ 0.10750e-3
						      (* y 0.74e-5)))))))))))))
       (let* ((y (/ 2.0 x)))
	 (* (/ (exp (- x)) (sqrt x)) 
	    (+ 1.25331414
	       (* y (+ -0.7832358e-1
		       (* y (+ 0.2189568e-1
			       (* y (+ -0.1062446e-1
				       (* y (+ 0.587872e-2
					       (* y (+ -0.251540e-2
						       (* y -0.53208e-3)))))))))))))))))


(defmacro bes-k1 (ux)				;K1(x)
  `(let ((x ,ux))
     (if (<= x 2.0)
	 (let* ((y (* x (/ x 4.0))))
	   (+ (* (log (/ x 2)) (bes-i1 x)) 
	      (* (/ 1.0 x)
		 (+ 1.0
		    (* y (+ 0.15443144
			    (* y (+ -0.67278579
				    (* y (+ -0.18156897
					    (* y (+ -0.1919402e-1
						    (* y (+ -0.110404e-2
							    (* y -0.4686e-4)))))))))))))))
       (let* ((y (/ 2.0 x)))
	 (* (/ (exp (- x)) (sqrt x)) 
	    (+ 1.25331414 
	       (* y (+ 0.23498619
		       (* y (+ -0.3655620e-1
			       (* y (+ 0.1504268e-1
				       (* y (+ -0.780353e-2
					       (* y (+ 0.325614e-2
						       (* y -0.68245e-3)))))))))))))))))


(defmacro bes-kn (un ux)			;return Kn(x) for any integer n, real x
  `(let ((n ,un) (x ,ux))
     (if (= n 0) (bes-k0 x)
       (if (= n 1) (bes-k1 x)
	 (do ((tox (/ 2.0 x))
	      (bkm (bes-k0 x))
	      (bk (bes-k1 x))
	      (bkp 0.0)
	      (j 1 (+ j 1)))
	     ((= j n) bk)
	   (setf bkp (+ bkm (* j tox bk)))
	   (setf bkm bk)
	   (setf bk bkp))))))


;;; Bessel-spectrum, see Lebedev p128, Tolstov p220, Watson chapter 18.
;;;
;;; a bes-jn spectrum requires knowlege of the zeros of Jn.  The weight function is x, the coeff weight is
;;;   xJn^2(lnx) summed over 0..1 = 2/J(n+1)^2(ln) (that is we need the value of J2 at the zero of J1 being used
;;;   in the nth function).

(defun bes-zero (n s &optional (j t))			;A&S 9.5.12 p371
  ;; checked through 40-th zero in Watson table vii p 748 for orders 0..5 
  (if (= s 0) 0.0
    (if (and j (= s 1) (= n 0)) 2.4048255577            ;these are particularly badly estimated
      (if (and (not j) (= s 1) (= n 0)) 0.89357697
      (let* ((B (* pi (+ s (if j -.25 -.75) (* .5 n))))
	     (B8 (* 8 B))
	     (B2 (* B8 B8))
	     (B3 (* 3 B8 B2))
	     (B5 (* 5 B3 B2))
	     (B7 (* 7 B5 B2))
	     (m (* 4 n n))
	     (m1 (- m 1))
	     (mm (* m m)))
	(- B 
	   (/ m1 B8) 
	   (/ (* 4 m1 (- (* 7 m) 31)) B3)
	   (/ (* 32 m1 (+ (* 83 mm) (* -982 m) 3779)) B5)
	   (/ (* 64 m1 (+ (* 6949 mm m) (* -153855 mm) (* 1585743 m) -6277237)) B7)))))))

(def-clm-fun 'bes-zero #'(lambda (var x) (package-op '<bes-zero> var x :clm-real)))

(defmacro <bes-zero> (result n s &optional (j t))
  (let ((lcr (lc (second result)))
	(lcs (lc-num-ref s))
	(lcn (lc-num-ref n))
	(lcj (lc-num-ref j)))
    (format *c-file* "  ~
      if (~A==0.0 ~A=0.0; else if ((~A==1) && (~A==0)) {if (~A == 0) ~A=0.89357697; else ~A=2.404825577;}~%  ~
      else {double b,b2,b8,b3,b5,b7,m,m1,mm; if (~A != 0) m=-.25; else m=-.75;~%     ~
            b=PI*(~A+m+(~A/2)); b8=8*b; b2=b8*b8; b3=3*b8*b2; b5=5*b3*b2; b7=7*b5*b2; m=4*~A*~A; m1=m-1; mm=m*m;~%     ~
            ~A=b-(m1/b8)-((4*m1*(7*m-31))/b3)-((32*m1*(83*mm-982*m+3779))/b5)-((64*m1*(6949*mm*m-153855*mm+1585743*m-6277237))/b7;}~%"
	    lcs lcr lcs lcn lcj lcr lcr
	    lcj
	    lcs lcn lcn lcn
	    lcr))
  nil)

;;; the bessel-spectrum deals in Jn and Yn only, and no guarantees for high n (the zero finder is
;;;   an approximation; it can be off by as much as .0002 even for low orders).

(defun bessel-spectrum (lim Fn samps An Zn order &optional y)
  ;; expand function in Fn as a sum of Bessel Functions (Jn if not y, else Yn)
  ;; coefficients (spectrum) returned in An, An and Zn are temporary working space
  ;; lim is how many Bessel zeros to use in the expansion -- should be a lot!
  ;; samps is how many samples to read in Fn
  ;; order is which Jn or Yn to use.  It is assumed here to be an integer.
  ;; Each function is formed by mapping 0..xn where xn is the nth zero of the order-th Jn|Yn into 0..1.
  ;; weight function is x, coeff weight is 2/(J(order+1)(xn)^2).  (dot product is x f(x->a..b) Jn(x->0..xn)).
  ;; The coefficients are O(1/(sqrt n)) (where n = zero tracker) (Watson 18.27 p595).
  ;;
  ;; Euler in 1764 used this to explain the vibrations of a circular membrane -- Watson chap 18 p576 and 1.3 p5.
  ;;   does that mean we can get a drum sound?  Butkov "Mathematical Physics" 9.7 shows the membrane and
  ;;   associated nodes. Yn are used for waves on a ring-shaped membrane (Butkov 9.10).  And modified Bessel
  ;;   for waves inside a circular cylinder.
  ;; 
  (loop for i from 0 to lim do (setf (aref An i) 0.0))
  (loop for i from 1 to lim do (setf (aref Zn i) (bes-zero order i (not y))))
  (let ((x 0.0d0)
	(rate (/ 1.0 samps)))
    (loop for k from 0 below samps do
      (let* ((fx (aref Fn k))
	     (xfx (* x fx)))
	(loop for i from 1 to lim do
	  (incf (aref An i) (* xfx (if y (bes-yn order (* x (aref Zn i))) 
				      (bes-jn order (* x (aref Zn i)))))))
	(incf x rate)))
    (let ((scaler (/ 2.0 samps)))
      (loop for i from 1 to lim do
	(let ((Jn1 (if y (bes-yn (1+ order) (aref Zn i))
		     (bes-jn (1+ order) (aref Zn i)))))
	  (setf (aref An i) (/ (* (aref An i) scaler) (* Jn1 Jn1)))))) ;probably wrong for Yn
    An))

;;; bessel-polynomial to parallel preceding spectra resynthesis
(defun bessel-polynomial (a x Zn order lim &optional y)
  (loop for i from 1 to lim do (setf (aref Zn i) (bes-zero order i (not y))))
  (let ((sum 0.0))
    (loop for i from 1 to lim do
      (incf sum (* (aref a i)
		   (if y (bes-yn order (* x (aref Zn i)))
		     (bes-jn order (* x (aref Zn i)))))))
    sum))


#|
;;; bes-i0 is used by asymmetric-fm for amplitude normalization -- see mus.lisp
;;; here we use J1 to try out Bessel-FM (i.e. J1(J1) essentially -- the bump at the
;;;   start gives us a percussive attack, and the damping of the sinusoid gives us
;;;   a dynamic spectrum.
;;; other Jn appear to give similar results.  Since the Bessel functions resemble a dampled sinusoid
;;;   after the start, the behaviour here is close to "normal" FM.  Bessel contrast-enhancement
;;;   should also resemble it's sine-based cousin. (Just use (* index (ina <file>)) as the argument to J1 below)

(definstrument bes-fm (beg dur freq amp ratio index)
  (let* ((st (floor (* beg *srate*)))
	 (nd (+ st (floor (* dur *srate*))))
	 (car-ph 0.0)
	 (mod-ph 0.0)
	 (car-incr (hz->radians freq))
	 (mod-incr (* ratio car-incr))
	 (ampenv (make-env :envelope '(0 0 25 1 75 1 100 0) :scaler amp :duration dur)))
    (run
     (loop for i from st to nd do
       (outa i (* (env ampenv) (bes-j1 car-ph)))
       (incf car-ph (+ car-incr (* index (bes-j1 mod-ph))))
       (incf mod-ph mod-incr)))))

;;; (with-sound (:statistics t) (bes-fm 0 1 440 10.0 1.0 4.0))
|#


;;; Airy functions (from A&S p446)
;;;   interesting because expressible in terms of Bessel functions of fractional order (1/3) and 
;;;   as x->-infinity, become similar to cos x^(3/2) (+inf->exponential) -- the exponent 3/2 gives a sweep.

(defun Airy (kind x)
  (if (> x 4.0)				;power series not good above this point
      (let ((chi (* 2/3 (expt x 1.5))))
	(if (= kind 0)
	    (* .5 (expt x -.25) (exp (- chi)) (+ .564190 (/ -.007 (- x 3.0))))
	  (* (expt x -.25) (exp chi) (+ .564190 (/ .014 (- x 3.0))))))
    (if (< x -10.0)			;or below this
	(let* ((xx (abs x))
	       (chi (* 2/3 (expt xx 1.5)))
	       (f1 (+ .3989423 (/ -.00144 (- xx 9.0))))
	       (f2 (+ .3989423 (/ .0013 (- xx 9.0)))))
	  (if (= kind 0)
	      (* (expt xx -.25) (+ (* (cos chi) f1) (* (sin chi) f2)))
	    (* (expt xx -.25) (- (* (cos chi) f2) (* (sin chi) f1)))))
      (let* ((c1 .355028053887817)
	     (c2 .258819403792807)
	     (err .000001)
	     (x3 (* x x x))
	     (sumf 1.0)
	     (unhappyf t)
	     (factf 1.0d0)
	     (xsf 1.0)
	     (fx (progn
		   (loop for i from 1 by 3 while unhappyf do
		     (setf xsf (* i xsf x3))
		     (loop for k from 1 to 3 do (setf xsf (/ xsf factf)) (incf factf))
		     (setf unhappyf (> (abs xsf) err))
		     (incf sumf xsf))
		   sumf))
	     (sumg x)
	     (factg 2.0d0)
	     (xsg x)
	     (unhappyg t)			
	     (gx (progn
		   (loop for i from 2 by 3 while unhappyg do
		     (setf xsg (* i xsg x3))
		     (loop for j from 1 to 3 do (setf xsg (/ xsg factg)) (incf factg))
		     (setf unhappyg (> (abs xsg) err))
		     (incf sumg xsg))
		   sumg)))
	(if (= kind 0)
	    (- (* c1 fx) (* c2 gx))
	  (* (sqrt 3.0) (+ (* c1 fx) (* c2 gx))))))))

(defmacro Ai (x) `(Airy 0 ,x))		;zeros at -2.33810 -4.087949 -5.520559 -6.786708 -7.944133 -9.0226508 -10.040174 (A&S p478)
(defmacro Bi (x) `(Airy 1 ,x))		;zeros at -1.17371 -3.271093 -4.830737 -6.169852 -7.376762 -8.4919488 -9.5381943

;;; the following Bessel functions could be computed about twice as fast in some cases by calculating the auxiliary
;;;   results just once.

(defmacro Jp3 (x)			;Jn where n=1/3
  `(let ((z (expt (* 1.5 ,x) .666667)))
     (* .5 (sqrt (/ 3 z)) (- (* (sqrt 3.0) (Ai (- z))) (Bi (- z))))))
(defmacro Jm3 (x)			;Jn when n=-1/3
  `(let ((z (expt (* 1.5 ,x) .666667)))
     (* .5 (sqrt (/ 3 z)) (+ (* (sqrt 3.0) (Ai (- z))) (Bi (- z))))))
(defmacro Ip3 (x)			;In n=1/3
  `(let ((z (expt (* 1.5 ,x) .666667)))
     (* .5 (sqrt (/ 3 z)) (- (Bi z) (* (sqrt 3.0) (Ai z))))))
(defmacro Im3 (x)			;In n=-1/3
  `(let ((z (expt (* 1.5 ,x) .666667)))
     (* .5 (sqrt (/ 3 z)) (+ (Bi z) (* (sqrt 3.0) (Ai z))))))

(def-clm-fun 'Airy #'(lambda (var x) (package-op '<Airy> var x :clm-real))) 

(defmacro <Airy> (result ukind ux)
  (let ((lcx (lc-num-ref ux))
	(lck (lc-num-ref ukind))
	(lcr (lc (second result))))
    (format *c-file* 
      "  {double x,c1,c2,err,x3,sumf,unhappyf,factf,xsf,fx,sumg,factg,xsg,unhappyg,gx,chi,f1,f2,xx; int i,k;~80,1T/* (~A ~A) */~%   ~
          x=~A;~%   ~
          if (x > 4.0) {chi=.666667*pow(x,1.5);~%      ~
            if (~A==0) {f1=.564190-(.007/(x-3.0)); ~A=.5*pow(x,-.25)*exp(-chi)*f1;}~%      ~
            else {f1=.564190+(.014/(x-3.0)); ~A=pow(x,-.25)*exp(chi)*f1;}}~%   ~
          else if (x < -10.0) {if (x<0.0) xx=-x; else xx=x; chi=.666667*pow(xx,1.5);~%      ~
            f1=.3989423-(.00144/(xx-9.0)); f2=.3989423+(.0013/(xx-9.0));~%      ~
            if (~A==0) ~A=pow(xx,-.25)*((f1*cos(chi))+(f2*sin(chi))); else ~A=pow(xx,-.25)*((f2*cos(chi))-(f1*sin(chi)));}~%      ~
            else {c1=.355028053887817; c2=.258819403792807; err=.000001; x3=x*x*x; sumf=1.0;unhappyf=1;~%   ~
              factf=1.0; xsf=1.0; sumg=x; factg=2.0; xsg=x; unhappyg=1; i=1;~%   ~
              while (unhappyf) {xsf=xsf*i*x3; for (k=1;k<=3;k++) {xsf=xsf/factf; factf+=1.0;}~%      ~
                                  if ((xsf<err) && (xsf>-err)) unhappyf=0; sumf+=xsf; i+=3;}~%   ~
                fx=sumf; i=2;~%   ~
                while (unhappyg) {xsg=xsg*i*x3; for (k=1;k<=3;k++) {xsg=xsg/factg; factg+=1.0;}~%      ~
                                  if ((xsg<err) && (xsg>-err)) unhappyg=0; sumg+=xsg; i+=3;}~%   ~
                gx=sumg; if (~A==0) ~A=c1*fx-c2*gx; else ~A=sqrt(3.0)*(c1*fx+c2*gx);}}~%"
	    (if (constantp ukind) (if (= ukind 0) "Ai" "Bi") (format nil "Airy ~A" ukind)) ux
	    lcx lck lcr lcr lck lcr lcr
	    lck lcr lcr)
    nil))

#|
(definstrument airy-x (beg dur amp freq)
  (let* ((st (floor (* *srate* beg)))
	 (nd (+ st (floor (* *srate* dur))))
	 (x 0.0)
	 (rate (/ freq *srate*)))
    (run
     (loop for i from st to nd do
       (outa i (* amp (Ai x)))
       (decf x rate)))))

;;; (with-sound () (airy-x 0 1 1 440.0))
|#


;;; sine and cosine integrals (also from A&S p230)

(defun aux-f (x)			;1<=x<inf
  (let ((x2 (* x x)))
    (/ (+ 38.102495 (* x2 (+ 335.677320 (* x2 (+ 265.187033 (* x2 (+ 38.027264 x2)))))))
       (* x (+ 157.105423 (* x2 (+ 570.236280 (* x2 (+ 322.624911 (* x2 (+ 40.021433 x2)))))))))))

(defun aux-g (x)
  (let ((x2 (* x x)))
    (/ (+ 21.821899 (* x2 (+ 352.018498 (* x2 (+ 302.757865 (* x2 (+ 42.242855 x2)))))))
       (* x2 (+ 449.690326 (* x2 (+ 1114.978885 (* x2 (+ 482.485984 (* x2 (+ 48.196927 x2)))))))))))

(defun Sint (x) 
  (if (>= x 1.0)
      (- (/ pi 2) (* (cos x) (aux-f x)) (* (sin x) (aux-g x)))
    (let* ((sum x)
	   (fact 2.0d0)
	   (one -1.0)
	   (xs x)
	   (x2 (* x x))
	   (err .000001)
	   (unhappy t))			
      (loop for i from 3.0 by 2.0 while unhappy do
	(setf xs (/ (* one x2 xs) (* i fact)))
	(setf one (- one))
	(incf fact)
	(setf xs (/ xs fact))
	(setf unhappy (> (abs xs) err))
	(incf sum xs))
      sum)))


(def-clm-fun 'sint #'(lambda (var x) (package-op '<sint> var x :clm-real))) 

(defmacro <sint> (result ux)
  (let ((lcx (lc-num-ref ux))
	(lcr (lc (second result))))
    (format *c-file* 
	    "{double x,x2,f,g,sum,fact,one,xs,err,unhappy,i; x=~A; x2=x*x;~80,1T/* (Sint ~A) */~%   ~
              if (x>=1.0) {~%      ~
                 f=((38.102495+(x2*(335.677320+(x2*(265.187033+(x2*(38.027264+x2)))))))/~%         ~
                    (x*(157.105423+(x2*(570.236280+(x2*(322.624911+(x2*(40.021433+x2)))))))));~%      ~
                 g=((21.821899+(x2*(352.018498+(x2*(302.757865+(x2*(42.242855+x2)))))))/~%         ~
                    (x2*(449.690326+(x2*(1114.978885+(x2*(482.485984+(x2*(48.196927+x2)))))))));~%      ~
                 ~A=PI/2.0 - (cos(x)*f + sin(x)*g);}~%   ~
              else {i=3.0; unhappy=1; sum=x; fact=2.0; one=-1.0; xs=x; err=.000001;~%      ~
                    while (unhappy) {xs=(xs*x2*one)/(i*fact); one=-one; fact+=1.0; xs=xs/fact; i+=2.0;~%         ~
                      if ((xs<err) && (xs>-err)) unhappy=0; sum+=xs;} ~A=sum;}}~%"
	    lcx ux
	    lcr lcr)
    nil))


(defun Cint (x) 
  (if (>= x 1.0)
      (- (* (sin x) (aux-f x)) (* (cos x) (aux-g x)))
    (let* ((g .5772156649)
	   (sum 0.0)
	   (fact 1.0d0)
	   (one -1.0)
	   (xs 1.0)
	   (x2 (* x x))
	   (err .000001)
	   (unhappy t))			
      (loop for i from 2.0 by 2.0 while unhappy do
	(setf xs (/ (* one x2 xs) (* i fact)))
	(setf one (- one))
	(incf fact)
	(setf xs (/ xs fact))
	(setf unhappy (> (abs xs) err))
	(incf sum xs))
      (+ g (log x) sum))))

(def-clm-fun 'cint #'(lambda (var x) (package-op '<cint> var x :clm-real))) 

(defmacro <cint> (result ux)
  (let ((lcx (lc-num-ref ux))
	(lcr (lc (second result))))
    (format *c-file* 
	    "{double x,x2,f,g,sum,fact,one,xs,err,unhappy,i; x=~A; x2=x*x;~80,1T/* (Cint ~A) */~%   ~
              if (x>=1.0) {~%      ~
                 f=((38.102495+(x2*(335.677320+(x2*(265.187033+(x2*(38.027264+x2)))))))/~%         ~
                    (x*(157.105423+(x2*(570.236280+(x2*(322.624911+(x2*(40.021433+x2)))))))));~%      ~
                 g=((21.821899+(x2*(352.018498+(x2*(302.757865+(x2*(42.242855+x2)))))))/~%         ~
                    (x2*(449.690326+(x2*(1114.978885+(x2*(482.485984+(x2*(48.196927+x2)))))))));~%      ~
                 ~A=((sin(x)*f) - (cos(x)*g));}~%   ~
              else {i=2.0; unhappy=1; sum=0.0; fact=1.0; one=-1.0; xs=1.0; err=.000001;~%      ~
                    while (unhappy) {xs=(xs*x2*one)/(i*fact); one=-one; fact+=1.0; xs=xs/fact; i+=2.0;~%         ~
                      if ((xs<err) && (xs>-err)) unhappy=0; sum+=xs;} ~A=.5772156649+log(x)+sum;}}~%"
	    lcx ux
	    lcr lcr)
    nil))


;;; gamma function's friends

(defmacro beta (uz uw)			;B(z,w)
  `(let ((z ,uz)
	 (w ,uw))
     (exp (- (+ (gammln z) (gammln w)) (gammln (+ z w))))))

(defmacro gser (ua ux)			;P(a,x) evaluated as series, also Ln(gamma)
  `(let ((a ,ua) (x ,ux))
     (if (minusp x) (error "~F is less than 0" x)
       (if (zerop x) 0.0
	 (do* ((gln (gammln a))
	       (itmax 100)
	       (eps 3.0e-7)
	       (ap a)
	       (sum (/ 1.0 a))
	       (del sum)
	       (n 1 (+ n 1)))
	     ((or (> n itmax)
		  (< (abs del) (* (abs sum) eps)))
	      (* sum (exp (+ (- x) (* a (log x)) (- gln)))))
	   (incf ap)
	   (setf del (* del (/ x ap)))
	   (incf sum del))))))

(defmacro gcf (ua ux)			;Q(a,x) evaluated as continued fraction
  `(let ((a ,ua) (x ,ux))
     (do* ((itmax 100)
	   (eps 3.0e-7)
	   (gln (gammln a))
	   (gold 0.0)			;previous value of g, tested for convergence
	   (a0 1.0)
	   (a1 x)
	   (b0 0.0)
	   (b1 1.0)			;setting up continued fraction
	   (fac 1.0)
	   (n 1 (+ n 1))
	   (ana 0.0) (g 0.0) (anf 0.0))
	 ((> n itmax) 
	  (* g (exp (+ (- x) (* a (log x)) (- gln)))))
       (setf ana (- n a))
       (setf a0 (* fac (+ a1 (* a0 ana))))
       (setf b0 (* fac (+ b1 (* b0 ana))))
       (setf anf (* fac n))
       (setf a1 (+ (* x a0) (* anf a1)))
       (setf b1 (+ (* x b0) (* anf b1)))
       (when (/= 0.0 a1)		;renormalize?
	 (setf fac (/ 1.0 a1))
	 (setf g (* b1 fac))
	 (if (< (abs (/ (- g gold) g)) eps)
	     (return (* g (exp (+ (- x) (* a (log x)) (- gln))))))))))

(defmacro gammp (ua ux)			;returns incomplete gamma function P(a,x)
  `(let ((a ,ua) (x ,ux))
     (if (or (<= a 0.0) (minusp x))
	 (error "Invalid argument to gammp: ~F" (if (<= 0.0 a) a x))
       (if (< x (+ a 1.0))
	   (gser a x)			;use series
	 (- 1.0 (gcf a x))))))		;use continued fraction
  
(defmacro gammq (a x)			;incomplete gamma function Q(a,x) = 1 - P(a,x)
  `(- 1.0 (gammp ,a ,x)))		;Numerical Recipes repeats the GAMMP procedure code -- kinda odd.

(defmacro erf (ux)			;returns error function erf(x)
  `(let ((x ,ux))
     (if (minusp x)
	 (- (gammp 0.5 (* x x)))
       (gammp 0.5 (* x x)))))

(defmacro erfc (ux)			;complementary error function erfc(x)
  `(let ((x ,ux))
     (if (minusp x)
	 (+ 1.0 (gammp 0.5 (* x x)))
       (gammq 0.5 (* x x)))))

(defun betacf (a b x)		;continued fraction for incomplete beta function ("Numerical Recipes" p188)
  (do* ((itmax 100)
	(eps 3.0e-7)
	(am 1.0) (bm 1.0) (az 1.0)
	(tem 0.0) (em 0.0) (d 0.0)
	(bpp 0.0) (bp 0.0) (app 0.0)
	(aold 0.0) (ap 0.0)
	(m 1 (+ m 1))
	(qab (+ a b))
	(qap (+ a 1.0))
	(qam (- a 1.0))
	(bz (- 1.0 (/ (* qab x) qap))))
      ((> m itmax) az)
    (setf em m)
    (setf tem (+ em em))
    (setf d (* em (- b m) (/ x (* (+ qam tem) (+ a tem)))))
    (setf ap (+ az (* d am)))
    (setf bp (+ bz (* d bm)))
    (setf d (* (- (+ a em)) (+ qab em) (/ x (* (+ a tem) (+ qap tem)))))
    (setf app (+ ap (* d az)))
    (setf bpp (+ bp (* d bz)))
    (setf aold az)
    (setf am (/ ap bpp))
    (setf bm (/ bp bpp))
    (setf az (/ app bpp))
    (setf bz 1.0)
    (if (< (abs (- az aold)) (* eps (abs az)))
	(return az))))

(def-clm-fun 'betacf #'(lambda (var x) (package-op '<betacf> var x :clm-real)))

(defmacro <betacf> (result a b x)
  (let ((lcr (lc (second result)))
	(lca (lc-num-ref a))
	(lcx (lc-num-ref x))
	(lcb (lc-num-ref b)))
    (format *c-file* "  ~
      {int happy,m; double eps,tem,qap,qam,qab,em,d,bz,bpp,bp,bm,az,app,am,aold,ap,paz,peps;  happy=1; eps=3.0e-7;~%  ~
       am=1.0; bm=1.0; az=1.0; qab=~A+~A; qap=~A+1.0; qam=~A-1.0; bz=1.0-qab*~A/qap; m=1;~%   ~
       while (happy) {em=m; tem=em+em; d=em*(~A-m)*~A/((qam+tem)*(~A+tem));~%     ~
       ap=az+d*am; bp=bz+d*bm; d=-(~A+em)*(qab+em*~A/((~A+tem)*(qap+tem)); app=ap+d*az; bpp=bp+d*bz;~%     ~
       aold=az; am=ap/bpp; bm=bp/bpp; az=app/bpp; bz=1.0; paz=az-aold; peps=eps*az;~%     ~
       if (peps<0.0) peps=-peps; if (paz<0.0) paz=-paz; if (paz<peps) happy=0; m++; if (m>=100) happy=0;} ~A=az;}~%"
	    lca lcb lca lca lcx 
	    lcb lcx lca 
	    lca lcx lca
	    lcr))
  nil)

(defmacro betai (ua ub ux)		;incomplete beta function Ix(a,b)
  `(let ((a ,ua) (b ,ub) (x ,ux))
     (if (or (minusp x) (> x 1.0))
	 (error "x is out of bounds: ~F" x)
       (let ((bt (if (or (zerop x) (= x 1.0)) 0.0
		   (exp (+ (gammln (+ a b))
			   (- (gammln a))
			   (- (gammln b))
			   (* a (log x))
			   (* b (log (- 1.0 x))))))))
	 (if (< x (/ (+ a 1.0) (+ a b 2.0)))
	     (* bt (/ (betacf a b x) a))
	   (- 1.0 (* bt (/ (betacf b a (- 1.0 x)) b))))))))

#|
(defun binomial (n m) 
  (let ((sum 1.0d0))
    (loop for nn from n downto (- n m -1) and mm = m then (1- mm) do
      (setf sum (/ (* sum nn) mm)))
    sum))
|#
(defmacro binomial (n m)
  `(exp (- (gammln (1+ ,n)) (+ (gammln (1+ ,m)) (gammln (- ,n ,m -1))))))


;;; -------- SPHERICAL BESSEL FUNCTIONS --------
;;;
;;; from "Numerical Library" p596 ff
;;;   for waves inside a spherical cavity (se Butkov 9.9)

(defun start (x n tt)
  (let ((s (- (* 2 tt) 1))
	(p (- (/ 36.0 x) tt))
	(r (/ n x))
	(q 0.0) (y 0.0))
    (if (or (> r 1.0) (= tt 1))
	(progn
	  (setf q (sqrt (+ (* r r) s)))
	  (setf r (- (* r (log (+ q r))) q)))
      (setf r 0.0))
    (setf q (+ (/ 18.0 x) r))
    (setf r (max p q))
    (setf p (sqrt (* 2.0 (+ tt r))))
    (setf p (* x (/ (+ 1.0 r p) (+ 1.0 p))))
    (setf q 0.0)
    (loop while (or (> p q) (< p (- q 1.0))) do
      (setf y p)
      (setf p (/ p x))
      (setf q (sqrt (+ (* p p) s)))
      (setf p (/ (* x (+ r q)) (log (+ p q))))
      (setf q y))
    (if (= tt 1) (floor (+ p 1.0)) (* -2 (floor (/ p -2.0))))))
    
    
(defun spher-bes-jn (x n Jn)
  (if (zerop x)
      (progn
	(setf (aref Jn 0) 1.0)
	(loop for i from n downto 1 do (setf (aref Jn i) 0.0)))
    (if (zerop n)
	(if (< (abs x) 0.015) 
	    (let ((x2 (/ (* x x) 6.0)))
	      (setf (aref Jn 0) (+ 1.0 (* x2 (- (* x2 0.3) 1.0)))))
	  (setf (aref Jn 0) (/ (sin x) x)))
      (let ((r 0.0))
	(loop for m from (start x n 0) downto 1 do
	  (setf r (/ 1.0 (- (/ (+ m m 1) x) r)))
	  (if (<= m n) (setf (aref Jn m) r)))
	(if (< x 0.015)
	    (let ((s (/ (* x x) 6.0)))
	      (setf r (+ (* s (- (* s 0.3) 1.0)) 1.0)))
	  (setf r (/ (sin x) x)))
	(setf (aref Jn 0) r)
	(loop for m from 1 to n do 
	  (setf (aref Jn m) (* (aref Jn m) r))
	  (setf r (aref Jn m)))))))

(defun spher-bes-yn (x n Yn)
  (setf (aref Yn 0) (- (/ (cos x) x)))
  (when (/= n 0)
    (let* ((yi2 (aref Yn 0))
	   (yi1 (/ (- yi2 (sin x)) x))
	   (yi 0.0))
      (setf (aref Yn 1) yi1)
      (loop for i from 2 to n do
	(setf yi (+ (- yi2) (/ (* (+ i i -1) yi1) x)))
	(setf (aref Yn i) yi)
	(setf yi2 yi1)
	(setf yi1 yi)))))

(defun nonexpspherbessi (x n In)
  (if (zerop x)
      (progn
	(setf (aref In 0) 1.0)
	(loop for i from n downto 1 do (setf (aref In i) 0.0)))
    (let ((x2 (+ x x))
	  (r 0.0))
      (setf x2 (if (< x2 0.7) (/ (sinh x) (* x (exp x))) (/ (- 1.0 (exp (- x2))) x2)))
      (setf (aref In 0) x2)
      (when (/= n 0)
	(loop for m from (start x n 1) downto 1 do
	  (setf r (/ 1.0 (+ r (/ (+ m m 1) x))))
	  (if (<= m n) (setf (aref In m) r)))
	(loop for m from 1 to n do 
	  (setf (aref In m) (* (aref In m) x2))
	  (setf x2 (aref In m)))))))

(defun spher-bes-in (x n In)
  (if (zerop x)
      (progn
	(setf (aref In 0) 1.0)
	(loop for i from n downto 1 do (setf (aref In i) 0.0)))
    (let ((expx (exp x)))
      (nonexpspherbessi x n In)
      (loop for i from n downto 0 do (setf (aref In i) (* (aref In i) expx))))))

(defun nonexpspherbessk (ux n Kn)
  (let* ((x (/ 1.0 ux))
	 (ki2 (* x pi .5)))
    (setf (aref Kn 0) ki2)
    (when (/= n 0)
      (let ((ki1 (* ki2 (+ 1.0 x))))
	(setf (aref Kn 1) ki1)
	(loop for i from 2 to n do
	  (let ((ki (+ ki2 (* (+ i i -1) x ki1))))
	    (setf (aref Kn i) ki)
	    (setf ki2 ki1)
	    (setf ki1 ki)))))))
	      
(defun spher-bes-kn (x n Kn)
  (let ((expx (exp (- x))))
    (nonexpspherbessk x n Kn)
    (loop for i from n downto 0 do (setf (aref Kn i) (* (aref Kn i) expx)))))

#|
;;; jinc:

(with-sound (:output "j")
  (loop for i from 0 below 1024 and r from 0.0 by (/ 1.0 1024) do
    (if (= r 0.0)
	(outa i .999)
      (outa i (/ (bes-j1 (* 40 pi r)) (* r 20 pi))))))
|#
